/******************************************************************************/
/*                                                                            */
/* Program name: MQXANCRD                                                      */
/* Author: James Davies                                                       */
/* Description:  C program for MQ coordinating XA-compliant database          */
/*               managers.                                                    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* Function:                                                                  */
/*                                                                            */
/*    MQXANCRD is a C program for MQ to coordinate getting a message off a     */
/*    queue and updating a databases within an MQSeries unit of work.         */
/*    This program calls functions that are in MQXADBNCD.SQC these contain     */
/*    the SQL to update the databases and hence must be prepared with the     */
/*    appropriate database.                                                   */
/*                                                                            */
/*    -- A message is read from a queue (under sync point), it must be in     */
/*       the form:                                                            */
/*                                                                            */
/*       UPDATE Balance change=nnn WHERE Account=nnn                          */
/*                                                                            */
/*       The sample AMQSPUT can be used to put the messages on the queue.     */
/*                                                                            */
/*    -- Information from the database is obtained and updated with the       */
/*       information in the message.                                          */
/*                                                                            */
/*    -- The new status of the database is printed.                           */
/*                                                                            */
/* Program logic:                                                             */
/*                                                                            */
/*    MQCONN connect to default queue manager (or optionally supplied name)   */
/*    MQOPEN open queue for input (using supplied parameter)                  */
/*    while no failures                                                       */
/*    .  MQBEGIN start a unit of work                                         */
/*    .  MQGET get next message from queue under sync point                   */
/*    .  get information from databases                                       */
/*    .  update information from databases                                    */
/*    .  MQCMIT commit changes                                                */
/*    .  print updated information                                            */
/*    .  (no message available counts as failure, and loop ends)              */
/*    MQCLOSE close queue                                                     */
/*    MQDISC disconnect from queue manager                                    */
/*                                                                            */
/******************************************************************************/
/*                                                                            */
/* MQXANCRD has 2 parameters - the name of the message queue (required)        */
/*                           - the queue manager name (optional)              */
/*                                                                            */
/******************************************************************************/

/******************************************************************************/
/*                                                                            */
/* Includes                                                                   */
/*                                                                            */
/******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <cmqc.h>                                /* MQI                       */

/******************************************************************************/
/*                                                                            */
/* Defines                                                                    */
/*                                                                            */
/******************************************************************************/

#define OK      0                                /* define OK as zero         */
#define NOT_OK  1                                /* define NOT_OK as one      */

/******************************************************************************/
/*                                                                            */
/* Function prototypes for SQL routines in MQXADBNCD.SQC                       */
/*                                                                            */
/******************************************************************************/

 int DeclareMQXADBCursor(void);
 int ConnectToMQXADB(void);
 int GetMQXATABDetails(char *name, long account, long *balance,
                        long *transactions);
 int UpdateMQXATABBalance(long balance, long transactions);


                                                 /*****************************/
int main(int argc, char *argv[])                 /* ----- START OF MAIN ----- */
                                                 /*****************************/
{
   /***************************************************************************/
   /* MQI structures                                                          */
   /***************************************************************************/
   MQOD od = {MQOD_DEFAULT};                     /* object descriptor         */
   MQOD odp = {MQOD_DEFAULT};                    /* object descriptor for put */
   MQMD md = {MQMD_DEFAULT};                     /* message descriptor        */
   MQMD mdp = {MQMD_DEFAULT};                    /* message descriptor for put*/
   MQGMO gmo = {MQGMO_DEFAULT};                  /* get message options       */
   MQPMO pmo = {MQPMO_DEFAULT};                  /* put message options       */
   MQBO bo = {MQBO_DEFAULT};                     /* begin options             */

   /***************************************************************************/
   /* MQI variables                                                           */
   /***************************************************************************/
   MQLONG rc=OK;                                 /* return code               */
   MQHCONN hCon;                                 /* handle to connection      */
   MQHOBJ hObj;                                  /* handle to object          */
   char QMName[50]="";                           /* default QM name           */
   MQLONG options;                               /* options                   */
   MQLONG reason;                                /* reason code               */
   MQLONG connReason;                            /* MQCONN reason code        */
   MQLONG compCode;                              /* completion code           */
   MQLONG openCompCode;                          /* MQOPEN completion code    */
   char msgBuf[5000];                             /* message buffer            */
   MQLONG msgBufLen;                             /* message buffer length     */
   MQLONG msgLen;                                /* message length received   */

   /***************************************************************************/
   /* Other variables                                                         */
   /***************************************************************************/
   char *pStr;                                   /* ptr to string             */
   long balanceChange;                           /* balance change            */
   char name[40];                                /* name                      */
   long account;                                 /* account number            */
   long balance;                                 /* balance                   */
   long transactions;                            /* transactions              */
   int gotMsg;                                   /* got message from queue    */
   int committedUpdate;                          /* committed update          */

   /***************************************************************************/
   /* First check we have been given correct arguments                        */
   /***************************************************************************/
   if (argc != 2 && argc != 3)
   {
      printf("Input is: %s 'queue name' 'queue manager name'.\n"
             "Note the queue manager name is optional\n", argv[0]);
      exit(99);
   } /* endif */

   if (argc == 3)                                /* use the queue manager     */
      strcpy(QMName, argv[2]);                   /* name supplied             */

   /***************************************************************************/
   /* Connect to queue manager                                                */
   /***************************************************************************/
   MQCONN(QMName, &hCon, &compCode, &connReason);

   if (compCode == MQCC_FAILED)
   {
     printf("MQCONN ended with reason code %li\n", connReason);
     exit((int) connReason);
   } /* endif */

   /***************************************************************************/
   /* Use input parameter as the name of the target queue                     */
   /***************************************************************************/
   strncpy(od.ObjectName, argv[1], (size_t) MQ_Q_NAME_LENGTH);
   printf("Target queue is %s\n", od.ObjectName);

   /***************************************************************************/
   /* Open the target message queue for input                                 */
   /***************************************************************************/
   options = MQOO_INPUT_AS_Q_DEF + MQOO_FAIL_IF_QUIESCING;

   MQOPEN(hCon, &od, options, &hObj, &openCompCode, &reason);

   if (reason != MQRC_NONE)
      printf("MQOPEN ended with reason code %li\n", reason);

   if (openCompCode == MQCC_FAILED)
   {
      printf("Unable to open queue for input\n");
      rc = openCompCode;                         /* stop further action       */
   } /* endif */

   /***************************************************************************/
   /* Declare the cursors for locking of reads from databases                 */
   /***************************************************************************/
   if (rc == OK)
      rc = DeclareMQXADBCursor();

   /***************************************************************************/
   /* Set up some things for the MQGET                                        */
   /***************************************************************************/
   msgBufLen = sizeof(msgBuf) - 1;
   gmo.Version = MQGMO_VERSION_2;    /* Avoid need to reset message ID and    */
   gmo.MatchOptions = MQMO_NONE;     /* correlation ID after every MQGET      */
   gmo.Options = MQGMO_WAIT + MQGMO_CONVERT + MQGMO_SYNCPOINT;
   gmo.WaitInterval = MQWI_UNLIMITED;         /* Unlimited limit for waiting  */

   /***************************************************************************/
   /* Get messages from the message queue, loop until there is a failure      */
   /***************************************************************************/
   while (compCode != MQCC_FAILED && rc == OK)
   {
      /************************************************************************/
      /* Set flags so that we can back out if something goes wrong and not    */
      /* lose the message.                                                    */
      /************************************************************************/
      gotMsg = 0;                                /* set flag to FALSE         */
      committedUpdate = 0;                       /* set flag to FALSE         */

      /************************************************************************/
      /* Start a unit of work                                                 */
      /************************************************************************/
      //MQBEGIN (hCon, &bo, &compCode, &reason);

      //if (reason == MQRC_NONE)
      //{
      //   printf("Unit of work started\n");
      //}
      //else
      //{
      //   printf("MQBEGIN ended with reason code %li\n", reason);
         /*********************************************************************/
         /* If we get a reason code and only a warning on the compCode, there */
         /* is something wrong with the resource manager so stop looping and  */
		 /* sort it out, whatever the compCode.                               */
         /*********************************************************************/
      //   rc = NOT_OK;                            /* stop looping              */
     // } /* endif */

      //if (compCode == MQCC_FAILED)
      //   printf("Unable to start a unit of work\n");

      /************************************************************************/
      /* Get message off queue                                                */
      /************************************************************************/
      if (rc == OK)
      {

         MQGET(hCon, hObj, &md, &gmo, msgBufLen, msgBuf, &msgLen,
               &compCode, &reason);

         if (reason != MQRC_NONE)
         {
            if (reason == MQRC_NO_MSG_AVAILABLE)
            {
               printf("No more messages\n");
            }
            else
            {
               printf("MQGET ended with reason code %li\n", reason);

               if (reason == MQRC_TRUNCATED_MSG_FAILED)
                  compCode = MQCC_FAILED;        /* stop looping              */
            } /* endif */
         }
         else
         {
            gotMsg = 1;                          /* set flag to TRUE          */
         } /* endif */
      } /* endif */

      /************************************************************************/
      /* Process the message received                                         */
      /************************************************************************/
      if (compCode != MQCC_FAILED && rc == OK)
      {
         msgBuf[msgLen] = '\0';                  /* add string terminator     */

         pStr = strstr(msgBuf, "UPDATE Balance change=");
         if (pStr != NULL)
         {
            pStr += sizeof("UPDATE Balance change=") -1;
            sscanf(pStr, "%li", &balanceChange);
         }
         else
         {
            printf("Invalid string received: %s\n", msgBuf);
            rc = NOT_OK;                         /* stop looping              */
         } /* endif */

         if (rc == OK)
         {
            pStr = strstr(msgBuf, "Account=");
            if (pStr != NULL)
            {
               pStr += sizeof("Account=") -1;
               sscanf(pStr, "%li", &account);
            }
            else
            {
               printf("Invalid string received: %s\n", msgBuf);
               rc = NOT_OK;                      /* stop looping              */
            } /* endif */
         } /* endif */

         /*********************************************************************/
         /* Get details from database                                         */
         /*********************************************************************/
         if (rc == OK)
            rc = ConnectToMQXADB();
         if (rc == OK)
            rc = GetMQXATABDetails(name, account, &balance, &transactions);


         /*********************************************************************/
         /* Update the bank balance                                           */
         /*********************************************************************/
         if (rc == OK)
         {
            transactions++;                      /* bump no of transactions   */
            balance += balanceChange;            /* alter balance             */

            if (rc == OK)
               rc = UpdateMQXATABBalance(balance, transactions);

            if (rc == OK)
            {
              // printf("Account No %li Balance updated from %li to %li %s\n",
              //        account, balance - balanceChange, balance, name);

               /******************************************************************/
               /*                                                                */
               /*   Use mqmd.replytoq and mqmd.replytoqmgr to reply              */
               /*                                                                */
               /******************************************************************/
               memset(&odp, '\0', sizeof(odp));
			   memcpy(odp.StrucId, MQOD_STRUC_ID, sizeof(MQOD_STRUC_ID)-1);
               odp.Version = MQOD_VERSION_1;
               odp.ObjectType = MQOT_Q;
               strncpy(odp.ObjectName, md.ReplyToQ, (size_t)MQ_Q_NAME_LENGTH);
               strncpy(odp.ObjectQMgrName, md.ReplyToQMgr, (size_t)MQ_Q_MGR_NAME_LENGTH);
              // printf("Reply queue is %s, Reply queue manager name is %s\n", odp.ObjectName, odp.ObjectQMgrName);

               /******************************************************************/
               /*                                                                */
               /*   Set up the pmo                                               */
               /*                                                                */
               /******************************************************************/
               memset(&pmo, '\0', sizeof(pmo));
               memcpy(pmo.StrucId, MQPMO_STRUC_ID, sizeof(MQPMO_STRUC_ID)-1);
               pmo.Version = MQPMO_VERSION_2;
               pmo.Options = MQPMO_SYNCPOINT;

               /**************************************************************/
               /* This is a well behaved MQ app                              */
               /* It puts the incoming msgid into the correlid of the reply  */
               /**************************************************************/
               //memset(&mdp, '\0', sizeof(mdp));
               memcpy(mdp.StrucId, MQMD_STRUC_ID, sizeof(MQMD_STRUC_ID)-1);
               mdp.Version = MQMD_VERSION_2;
               mdp.Version = 2;
               mdp.Report = MQRO_NONE;
               mdp.MsgType = MQMT_DATAGRAM;
               memcpy(mdp.Format, "MQSTR   ", 8);
               mdp.Persistence = MQPER_PERSISTENT;
               mdp.MsgFlags = MQMF_NONE;
			   mdp.Expiry = MQEI_UNLIMITED;
               memcpy(mdp.CorrelId, md.MsgId, sizeof(md.MsgId) );
               memcpy(mdp.MsgId, MQMI_NONE, sizeof(md.MsgId) );

               MQPUT1(hCon,                /* connection handle               */
                      &odp,                /* object descriptor               */
                      &mdp,                /* message descriptor              */
                      &pmo,                /* default options (datagram)      */
                      msgLen,              /* message length                  */
                      msgBuf,              /* message buffer                  */
                      &compCode,           /* completion code                 */
                      &reason);            /* reason code                     */
                  
               /* report reason, if any */
               if (reason != MQRC_NONE)
               {
                 printf("MQPUT1 ended with reason code %ld\n", reason);
                 rc = NOT_OK;
               }
               else
               {

                 /***************************************************************/
                 /* We are going to commit the update so even if something goes */
                 /* wrong now, the message has been used so don't back out.     */
                 /***************************************************************/
                 committedUpdate = 1;              /* set flag to TRUE          */

                 /***************************************************************/
                 /* Note: the cursor will be implicitly closed by the MQCMIT.   */
                 /***************************************************************/
                 MQCMIT(hCon, &compCode, &reason);

                 if (reason == MQRC_NONE)
                 {
                //    printf("Unit of work successfully completed\n");
                 }
                 else
                 {
                    printf("MQCMIT ended with reason code %li completion code "
                           "%li\n", reason, compCode);
                    rc = NOT_OK;                   /* stop looping              */
                 } /* endif */
              } /* endif */
            } /* endif */
         } /* endif */
      } /* endif */

      /************************************************************************/
      /* If we got the message but something went wrong, back out so that we  */
      /* don't lose the message.                                              */
      /************************************************************************/
      if (gotMsg && !committedUpdate)
      {
         MQBACK(hCon, &compCode, &reason);

         if (reason == MQRC_NONE)
            printf("MQBACK successfully issued\n");
         else
            printf("MQBACK ended with reason code %li\n", reason);
      } /* endif */
   } /* endwhile */

   /***************************************************************************/
   /* Close queue if opened                                                   */
   /***************************************************************************/
   if (openCompCode != MQCC_FAILED)
   {
      options = 0;                               /* no close options          */

      MQCLOSE(hCon, &hObj, options, &compCode, &reason);

      if (reason != MQRC_NONE)
         printf("MQCLOSE ended with reason code %li\n", reason);
   } /* endif */

   /***************************************************************************/
   /* Disconnect from queue manager if not already disconnected               */
   /***************************************************************************/
   if (connReason != MQRC_ALREADY_CONNECTED)
   {
      MQDISC(&hCon, &compCode, &reason);

      if (reason != MQRC_NONE)
         printf("MQDISC ended with reason code %li\n", reason);
   } /* endif */

   return 0;
                                                 /*****************************/
}                                                /* ------ END OF MAIN ------ */
                                                 /*****************************/
